#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
# matopt.materials.parsers

# TODO: Decide if (and how) we should support:
# from matopt.materials.parsers import *
# NOTE: Probably will never need this, since parsers should
#       probably just be incorporated in Design.py and most
#       of these methods should be behind-the-scenes.

'''
import matopt.materials.parsers.CFG
import matopt.materials.parsers.PDB
import matopt.materials.parsers.POSCAR
import matopt.materials.parsers.XYZ
'''
# TODO: Decide if this is necessary
